<?php

if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

/**
 * Class Associator
 */
class Associator
{
    // Unique provider UUID in AssociatorAPI
    const PROVIDER = '07fb7006-773a-4c4e-82dd-6f9dc05cbaeb';
    const EXPORT_FIELD_NAME = 'associator_export';
    const REPORT_FIELD_NAME = 'associator_report';
    const DEFAULT_MAX_RECOMMENDATIONS = 3;

    /** @var \Associator\Associator */
    private $associator;

    /** @var Associator_Event_Repository */
    private $eventRepository;

    /**
     * Associator constructor.
     * @param $associator
     * @param $eventRepository
     */
    public function __construct($associator, $eventRepository)
    {
        $this->associator = $associator;
        $this->eventRepository = $eventRepository;
    }

    /**
	 * Initialize
	 */
	public function init()
    {
        $this->register_action();
        $this->register_shortcode();
        $this->display_options();
	}

    /**
     * Register actions
     */
	public function register_action()
    {
        add_action('init',[$this, 'register_session']);
        add_action('wp_ajax_associator_create_application', [$this, 'create_application']);
        add_action('wp_ajax_associator_import_transactions', [$this, 'import_transactions']);
        add_action('wp_ajax_associator_report', [$this, 'report']);
        add_action('wp_ajax_associator_ajax_add_to_cart', [$this, 'ajax_add_to_cart']);
        add_action('wp_ajax_nopriv_associator_ajax_add_to_cart', [$this, 'ajax_add_to_cart']);
        add_action('wp_ajax_associator_ajax_event', [$this, 'ajax_event']);
        add_action('wp_ajax_nopriv_associator_ajax_event', [$this, 'ajax_event']);
        add_action('woocommerce_thankyou', [$this, 'order_status_completed']);
    }

    /**
     * Register shortcode
     */
    public function register_shortcode()
    {
        add_shortcode('associator', [$this, 'associator_shortcode']);
    }

    /**
     * Handle display options action
     */
    public function display_options()
    {
        $settings = get_option('associator_settings', []);

        if (isset($settings['associator_single_product_recommendations']) && boolval($settings['associator_single_product_recommendations'])) {
            add_action( 'woocommerce_after_single_product_summary', function () {
                $title = __( 'Customers Who Bought This Item Also Bought', 'associator' );
                $shortcode = sprintf('[associator max="%d"]%s[/associator]', self::DEFAULT_MAX_RECOMMENDATIONS, $title);
                echo do_shortcode(apply_filters( 'widget_text', $shortcode));
            }, 20 );
        }

        if (isset($settings['associator_hide_related_products']) && boolval($settings['associator_hide_related_products'])) {
            remove_action( 'woocommerce_after_single_product_summary', 'woocommerce_output_related_products', 20 );
        }

        if (isset($settings['associator_cart_recommendations']) && boolval($settings['associator_cart_recommendations'])) {
            add_action( 'woocommerce_cart_collaterals', function () {
                $title = __( 'You may be interested in&hellip;', 'associator' );
                $shortcode = sprintf('[associator max="%d"]%s[/associator]', self::DEFAULT_MAX_RECOMMENDATIONS, $title);
                echo do_shortcode(apply_filters( 'widget_text', $shortcode));
            }, 20 );
        }

        if (isset($settings['associator_hide_cross_sells']) && boolval($settings['associator_hide_cross_sells'])) {
            remove_action( 'woocommerce_cart_collaterals', 'woocommerce_cross_sell_display' );
        }
    }

    /**
     * Register session if not start already
     */
    public function register_session()
    {
        if(!session_id()) {
            session_start();
        }
    }

    /**
     * Get products form basket
     * @return array
     */
    public function get_products_from_cart()
    {
        $products = [];

        foreach (WC()->cart->get_cart() as $cartItem => $values) {
            if ($values['quantity'] > 0) {
                $products[] = $values['product_id'];
            }
        }

        return $products;
    }

    /**
     * Add products to basket
     */
    function ajax_add_to_cart()
    {
        if (!session_id()) {
            session_start();
        }

        $products = isset($_POST['products']) && is_array($_POST['products']) ? $_POST['products'] : [];
        $_SESSION['associator'] = isset($_SESSION['associator']) ? array_merge($_SESSION['associator'], $products) : $products;

        $this->eventRepository->persist(
            Associator_Event_Repository::EVENT_TYPE_ADD,
            array_map(function ($product) { return intval($product['product_id']); }, $products)
        );

        foreach ($products as $product) {
            $product_id = absint($product['product_id']);
            $variation_id = absint($product['variation_id']);
            WC()->cart->add_to_cart($product_id, 1, $variation_id);
        }

        WC_AJAX :: get_refreshed_fragments();
    }

    /**
     * Notify event
     */
    function ajax_event()
    {
        $event = isset($_POST['event']) ? $_POST['event'] : null;
        $products = isset($_POST['products']) && is_array($_POST['products']) ? $_POST['products'] : [];

        if (!in_array($event, $this->eventRepository->getAllowedEvents())) {
            return;
        }

        $this->eventRepository->persist($event, array_map('intval', $products));
    }

    /**
     * Render recommendations using short code
     * @param $attributes
     * @param null $content
     * @return null|string
     */
    function associator_shortcode($attributes, $content = '')
    {
        $attributes = shortcode_atts(['max' => self::DEFAULT_MAX_RECOMMENDATIONS], $attributes);
        $settings = get_option('associator_settings', []);

        if (!isset($settings['associator_api_key'])) {
            return null;
        }

        // Get products from basket
        $samples = $this->get_products_from_cart();

        // Return if basket is empty
        if (empty($samples)) {
            return null;
        }

        $this->associator->setApiKey($settings['associator_api_key']);
        //Get support and confidence and normalize
        $support = isset($settings['associator_support']) ? floatval($settings['associator_support']) / 100 : null;
        $confidence = isset($settings['associator_confidence']) ? floatval($settings['associator_confidence']) / 100 : null;

        try {
            $response = $this->associator->getAssociations($samples, $support, $confidence);
        } catch (\Associator\Exception\AssociatorException $exception) {
            return null;
        }
        if (isset($response['status']) && $response['status'] === 'Error') {
            return null;
        }

        $associations = isset($response['associations']) ? $response['associations'] : [];
        $associations = array_reduce($associations, 'array_merge', array());

        // Return if no associations found
        if (empty($associations)) {
            return null;
        }

        $products = [];
        foreach ($associations as $association) {
            if (wc_get_product($association) === false) {
                continue;
            }

            if (count($products) >= $attributes['max']) {
                break;
            }

            $products[] = $association;
        }

        // Return if recommendations is empty
        if (empty($products)) {
            return null;
        }

        $this->eventRepository->persist(Associator_Event_Repository::EVENT_TYPE_SHOW, $products);

        ob_start();

        remove_action( 'woocommerce_after_shop_loop_item', 'woocommerce_template_loop_add_to_cart', 10 );
        remove_action( 'woocommerce_shop_loop_item_title','woocommerce_template_loop_product_title', 10 );
        add_action( 'woocommerce_shop_loop_item_title', [$this, 'loop_item_title_template'], 10 );
        add_action( 'woocommerce_shortcode_after_products_loop', [$this, 'after_products_loop_template'], 10 );

        include('views/short-code.php');

        remove_action( 'woocommerce_shop_loop_item_title', [$this, 'loop_item_title_template'], 10 );
        remove_action( 'woocommerce_shortcode_after_products_loop', [$this, 'after_products_loop_template'], 10 );
        add_action( 'woocommerce_after_shop_loop_item', 'woocommerce_template_loop_add_to_cart', 10 );
        add_action( 'woocommerce_shop_loop_item_title','woocommerce_template_loop_product_title', 10 );

        return ob_get_clean();
    }

    /**
     * Order status completed action
     * @param $order_id
     * @return bool
     */
    function order_status_completed($order_id)
    {
        $settings = get_option('associator_settings', []);

        if (!isset($settings['associator_api_key'])) {
            return false;
        }

        $this->associator->setApiKey($settings['associator_api_key']);
        $order = new WC_Order($order_id);

        // Prepare array with product ids from order
        $items = array_map(function ($item) { return $item->get_product_id(); }, $order->get_items());
        try {
            $response = $this->associator->saveTransaction(array_values($items));
        } catch (\Associator\Exception\AssociatorException $exception) {
            return false;
        }

        // Add information about persist in Associator
        if (isset($response['status']) && $response['status'] === 'Success') {
            add_post_meta($order->get_id(), self::EXPORT_FIELD_NAME, $settings['associator_api_key']);
        }

        if (!isset($_SESSION['associator'])) {
            return false;
        }

        foreach ($_SESSION['associator'] as $product) {
            if (in_array($product['product_id'], $items)) {
                $report[] = intval($product['product_id']);
            }
        }

        $this->eventRepository->persist(
            Associator_Event_Repository::EVENT_TYPE_BUY,
            array_unique($report)
        );

        unset($_SESSION['associator']);
    }

    /**
     * Create application in AssociatorAPI
     */
	public function create_application()
    {
        $name = parse_url(get_site_url(), PHP_URL_HOST);
        $response = $this->associator->createApplication($name, self::PROVIDER);

        if (isset($response['status']) && $response['status'] === 'Error') {
            wp_send_json(['status' => 'Error', 'message' => $response['message']]);
        }

        $settings = get_option('associator_settings', []);
        $settings['associator_api_key'] = $response['id'];
        update_option('associator_settings', $settings);

        wp_send_json(['status' => 'Success', 'message' => 'Application created successfully.']);
    }

    /**
     * Get reports data
     */
    public function report()
    {
        $from = new DateTime($_GET['from']);
        $to = new DateTime($_GET['to']);
        $interval = new \DateInterval('P1D');
        $period = new \DatePeriod($from, $interval, $to);

        foreach ($period as $key => $value) {

            // DATE LABELS
            $label[] = $value->format('d-m');

            // SHOW SERIES
            $events = $this->eventRepository->getByEventAndDate(Associator_Event_Repository::EVENT_TYPE_SHOW, $value->format('Y-m-d'));
            $items = array_map(function ($event) { return count(json_decode($event['value']));}, $events);
            $showSeries[] = array_sum($items);

            // ADD SERIES
            $events = $this->eventRepository->getByEventAndDate(Associator_Event_Repository::EVENT_TYPE_ADD, $value->format('Y-m-d'));
            $items = array_map(function ($event) { return count(json_decode($event['value']));}, $events);
            $addSeries[] = array_sum($items);

            // BUY SERIES
            $events = $this->eventRepository->getByEventAndDate(Associator_Event_Repository::EVENT_TYPE_BUY, $value->format('Y-m-d'));
            $items = array_map(function ($event) { return count(json_decode($event['value']));}, $events);
            $buySeries[] = array_sum($items);

            // CLICK SERIES
            $events = $this->eventRepository->getByEventAndDate(Associator_Event_Repository::EVENT_TYPE_CLICK, $value->format('Y-m-d'));
            $items = array_map(function ($event) { return count(json_decode($event['value']));}, $events);
            $clickSeries[] = array_sum($items);

            // ORDER SERIES
            $orders = wc_get_orders(['date_created' => $value->format('Y-m-d'), 'limit' => -1]);
            $count = array_map(function ($order) { return count($order->get_items()); }, $orders);
            $orderSeries[] = array_sum($count);
        }

        wp_send_json([
            'status' => 'Success',
            'charts' => [
                'orders' => [
                    'labels' => $label,
                    'series' => [$orderSeries],
                ],
                'views' => [
                    'labels' => $label,
                    'series' => [$showSeries],
                ],
                'add' => [
                    'labels' => $label,
                    'series' => [$addSeries],
                ],
                'click' => [
                    'labels' => $label,
                    'series' => [$clickSeries],
                ],
                'buy' => [
                    'labels' => $label,
                    'series' => [$buySeries],
                ]
            ]
        ]);
    }

    /**
     * Prepare import data with transactions and send to AssociatorAPI
     */
    public function import_transactions()
    {
        $transactions = [];
        $orders = [];
        $settings = get_option('associator_settings', []);

        if (!isset($settings['associator_api_key'])) {
            wp_send_json(['status' => 'Error', 'message' => 'Api key must be set.']);
        }

        $this->associator->setApiKey($settings['associator_api_key']);

        foreach (wc_get_orders(['limit' => -1]) as $order) {

            // Not import order if already persist in Associator
            if (in_array($settings['associator_api_key'], get_post_meta($order->get_id(), self::EXPORT_FIELD_NAME))) {
                continue;
            }

            // Prepare array witch transactions
            $transactions[] = array_map(function ($item) { return $item->get_product_id(); }, $order->get_items());
            $orders[] = $order->get_id();
        }

        if (empty($transactions)) {
            wp_send_json(['status' => 'Success', 'message' => 'Transaction imported.']);
        }

        try {
            $response = $this->associator->importTransactions($transactions);
        } catch (\Associator\Exception\AssociatorException $exception) {
            wp_send_json(['status' => 'Error', 'message' => $exception->getMessage()]);
        }

        if (isset($response['status']) && $response['status'] === 'Error') {
            wp_send_json(['status' => 'Error', 'message' => $response['message']]);
        }

        // Add information about persist in Associator
        foreach ($orders as $order) {
            add_post_meta($order, self::EXPORT_FIELD_NAME, $settings['associator_api_key']);
        }

		update_option( 'associator_orders_exported', 1 );

        wp_send_json(['status' => 'Success', 'message' => 'Transaction imported.']);
    }

    public function loop_item_title_template()
    {
        echo '<div class="associator-checkbox-add-to-cart"><label><h2><input type="checkbox" class="woocommerce-form__input woocommerce-form__input-checkbox input-checkbox" value="' . get_the_id() . '" checked> <span class="woocommerce-loop-product__title">' . get_the_title() . '</span></h2></label></div>';
    }

    public function after_products_loop_template()
    {
        echo '<div class="associator-button-add-to-cart"><a id="associator_add_to_card" class="single_add_to_cart_button button also_bought_css_button" href="#">Add to the basket</a></div>';
    }
}
